# -*- coding: utf-8 -*-
__all__ = ('collect_functions',)

from typing import Iterator, Collection, List
import ast

from loguru import logger
import asttokens
import astor

from .analysis import PythonFunction
from .util import ast_with_tokens, ast_location
from ..container import ProjectContainer
from ..core import Location, FileLocationRange, LocationRange
from ..functions import ProgramFunctions


def collect_functions(container: ProjectContainer) -> ProgramFunctions:
    """Finds all functions within a Python project given a container."""
    logger.debug(f'collecting functions for project [{container.project}]')
    visitor = CollectFunctionsVisitor(container)
    for filename in container.project.files:
        visitor.collect(filename)
    return ProgramFunctions(visitor.functions)


class CollectFunctionsVisitor(ast.NodeVisitor):
    def __init__(self, container: ProjectContainer) -> None:
        super().__init__()
        self.atok: asttokens.ASTTokens
        self.container = container
        self.functions: List[PythonFunction] = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        location = ast_location(self.atok, node)
        body_location = ast_location(self.atok, node.body)
        function = PythonFunction(name=node.name,
                                  location=location,
                                  body_location=body_location)
        logger.debug(f"found function definition: {function}")
        self.functions.append(function)

    def collect(self, filename: str) -> None:
        self.atok = ast_with_tokens(self.container, filename)
        project = self.container
        logger.debug(f'collecting functions in file {filename} '
                     f'for project [{project}]')
        self.visit(self.atok.tree)



# -*- coding: utf-8 -*-
__all__ = ('ProjectContainer',)

import typing

import attr
import dockerblade as _dockerblade

if typing.TYPE_CHECKING:
    from .project import Project


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ProjectContainer:
    """Provides access to a container for a project under analysis.

    Attributes
    ----------
    project: Project
        The project under analysis.
    dockerblade: dockerblade.Container
        The underlying DockerBlade container for the project.
    shell: dockerblade.Shell
        Provides access to an executable shell for this container.
    files: dockerblade.FileSystem
        Provides root-level access to the filesystem for this container.
    """
    project: 'Project'
    dockerblade: _dockerblade.Container
    shell: _dockerblade.Shell = attr.ib(init=False)
    files: _dockerblade.FileSystem = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        object.__setattr__(self, 'shell', self.dockerblade.shell('/bin/sh'))
        object.__setattr__(self, 'files', self.dockerblade.filesystem())



# -*- coding: utf-8 -*-
__all__ = ('collect_loops',)

from typing import Union, List
import ast

from loguru import logger
import asttokens
import astor

from .util import ast_with_tokens, ast_location
from ..container import ProjectContainer
from ..core import FileLocationRange
from ..loops import ProgramLoops


def collect_loops(container: ProjectContainer) -> ProgramLoops:
    """Finds all loops within a Python project given a container."""
    logger.debug(f'collecting loops for project [{container.project}]')
    visitor = CollectLoopsVisitor(container)
    for filename in container.project.files:
        visitor.collect(filename)
    return ProgramLoops.from_body_location_ranges(visitor.locations)


class CollectLoopsVisitor(ast.NodeVisitor):
    def __init__(self, container: ProjectContainer) -> None:
        super().__init__()
        self.atok: asttokens.ASTTokens
        self.container = container
        self.locations: List[FileLocationRange] = []

    def visit_loop(self,
                   node: Union[ast.For, ast.AsyncFor, ast.While]
                   ) -> None:
        self.locations.append(ast_location(self.atok, node.body))
        if node.orelse:
            self.locations.append(ast_location(self.atok, node.orelse))

    visit_For = visit_loop
    visit_AsyncFor = visit_loop
    visit_While = visit_loop

    def collect(self, filename: str) -> None:
        self.atok = ast_with_tokens(self.container, filename)
        project = self.container
        logger.debug(f'collecting loops in file {filename} '
                     f'for project [{project}]')
        self.visit(self.atok.tree)



# -*- coding: utf-8 -*-
from .version import __version__
from .exceptions import KaskaraException
from .project import Project
from .analysis import Analysis
from .analyser import Analyser
from .insertions import InsertionPoint
from .statements import Statement
from .loops import ProgramLoops

from . import clang
from . import python



# -*- coding: utf-8 -*-
import pytest

import os

import docker as _docker
import dockerblade as _dockerblade
import kaskara

from kaskara.core import FileLine
from kaskara.statements import ProgramStatements
from kaskara.functions import ProgramFunctions
from kaskara.loops import ProgramLoops
from kaskara.python import PythonStatement

DIR_HERE = os.path.dirname(__file__)


@pytest.fixture
def flask():
    docker = _docker.from_env()
    docker_image_name = 'kaskara/examples:flask'
    docker_directory = os.path.join(DIR_HERE, 'examples/flask')

    # build the Docker image
    docker.images.build(path=docker_directory, tag=docker_image_name, rm=True)

    with _dockerblade.DockerDaemon() as dockerblade:
        files = {
            'flask/signals.py',
            'flask/helpers.py'
        }
        yield kaskara.Project(dockerblade=dockerblade,
                              image=docker_image_name,
                              directory='/opt/flask/src',
                              files=files)


def test_collect_statements(flask):
    def line(num: int) -> FileLine:
        return FileLine('flask/signals.py', num)

    with flask.provision() as container:
        filename = 'flask/signals.py'
        visitor = kaskara.python.statements.CollectStatementsVisitor(container)
        visitor.collect(filename)
        statements = ProgramStatements(visitor.statements)

        statements_at_line = list(statements.at_line(line(21)))
        assert len(statements_at_line) == 1
        actual = statements_at_line[0]
        assert actual.content == 'return _FakeSignal(name, doc)'


def test_collect_functions(flask):
    with flask.provision() as container:
        filename = 'flask/signals.py'
        visitor = kaskara.python.functions.CollectFunctionsVisitor(container)
        visitor.collect(filename)
        functions = ProgramFunctions(visitor.functions)

        assert len(functions) == 4
        assert set(f.name for f in functions) == {'signal',
                                                  '__init__',
                                                  'send',
                                                  '_fail'}


def test_collect_loops(flask):
    with flask.provision() as container:
        filename = 'flask/helpers.py'
        visitor = kaskara.python.loops.CollectLoopsVisitor(container)
        visitor.collect(filename)
        loops = ProgramLoops.from_body_location_ranges(visitor.locations)
        body_locations = list(loops._covered_by_loop_bodies)

        assert len(body_locations) == 3



# -*- coding: utf-8 -*-
from sourcelocation import (Location,
                            LocationRange,
                            FileLocation,
                            FileLocationRange,
                            FileLocationRangeSet,
                            FileLine)



# -*- coding: utf-8 -*-
__all__ = ('collect_statements',)

from typing import Iterator, Collection, List
import ast
import asttokens

import astor
from loguru import logger

from .analysis import PythonStatement
from .util import ast_with_tokens, ast_location
from ..container import ProjectContainer
from ..core import Location, FileLocationRange, LocationRange
from ..statements import Statement, ProgramStatements

STMT_CLASS_NAMES = {
    'FunctionDef',
    'AsyncFunctionDef',
    'ClassDef',
    'Return',
    'Delete',
    'Assign',
    'AugAssign',
    'AnnAssign',
    'For',
    'AsyncFor',
    'While',
    'If',
    'With',
    'AsyncWith',
    'Raise',
    'Try',
    'Assert',
    'Import',
    'ImportFrom',
    'Global',
    'Nonlocal',
    'Expr',
    'Pass',
    'Break',
    'Continue'
}


def collect_statements(container: ProjectContainer) -> ProgramStatements:
    """Finds all statements within a Python project given a container."""
    logger.debug(f'collecting statements for project [{container.project}]')
    visitor = CollectStatementsVisitor(container)
    for filename in container.project.files:
        visitor.collect(filename)
    return ProgramStatements(visitor.statements)


class CollectStatementsVisitor(ast.NodeVisitor):
    def __init__(self, container: ProjectContainer) -> None:
        super().__init__()
        self.atok: asttokens.ASTTokens
        self.container = container
        self.statements: List[PythonStatement] = []

    def visit_Module(self, node: ast.Module) -> None:
        for stmt in node.body:
            self.visit(stmt)

    def visit_ExceptHandler(self, node: ast.ExceptHandler) -> None:
        for stmt in node.body:
            self.visit(stmt)

    def generic_visit(self, node: ast.AST) -> None:
        logger.debug(f"visiting node: {node}")
        if node.__class__.__name__ not in STMT_CLASS_NAMES:
            return
        self._collect_stmt(node)
        super().generic_visit(node)

    def _collect_stmt(self, node: ast.AST) -> None:
        kind = node.__class__.__name__
        canonical = astor.to_source(node)
        location = ast_location(self.atok, node)
        source = self.atok.get_text(node)
        stmt = PythonStatement(kind=kind,
                               content=source,
                               canonical=canonical,
                               location=location)
        logger.debug(f"found statement: {stmt}")
        logger.debug(f"statement at location: {location}")
        self.statements.append(stmt)

    def collect(self, filename: str) -> None:
        self.atok = ast_with_tokens(self.container, filename)
        project = self.container
        logger.debug(f'collecting statements in file {filename} '
                     f'for project [{project}]')
        self.visit(self.atok.tree)



# -*- coding: utf-8 -*-
__all__ = ('Analysis',)

import attr

from .core import FileLocation
from .functions import ProgramFunctions
from .insertions import ProgramInsertionPoints
from .loops import ProgramLoops
from .project import Project
from .statements import ProgramStatements


@attr.s(slots=True, auto_attribs=True, frozen=True)
class Analysis:
    """Stores the results of an analysis of a given project.

    Attributes
    ----------
    project: Project
        A description of the project that was analysed.
    loops: ProgramLoops
        The set of loop control-flow statements within the program.
    functions: ProgramFunctions
        The set of functions within the program.
    statements: ProgramStatements
        The set of statements within the program.
    """
    project: Project
    loops: ProgramLoops
    functions: ProgramFunctions
    statements: ProgramStatements
    insertions: ProgramInsertionPoints

    def is_inside_loop(self, location: FileLocation) -> bool:
        return self.loops.is_within_loop(location)

    def is_inside_function(self, location: FileLocation) -> bool:
        return self.functions.encloses(location) is not None

    def is_inside_void_function(self, location: FileLocation) -> bool:
        f = self.functions.encloses(location)
        return f is not None and f.return_type == 'void'



# -*- coding: utf-8 -*-
__all__ = ('PythonFunction', 'PythonStatement')

from typing import Sequence, Optional, FrozenSet

import attr

from ..core import FileLocationRange
from ..functions import Function
from ..statements import Statement


@attr.s(frozen=True, auto_attribs=True, slots=True)
class PythonFunction(Function):
    name: str
    location: FileLocationRange
    body_location: FileLocationRange

    @property
    def return_type(self) -> Optional[str]:
        return None


@attr.s(frozen=True, auto_attribs=True, slots=True)
class PythonStatement(Statement):
    kind: str
    content: str
    canonical: str
    location: FileLocationRange

    @property
    def visible(self) -> Optional[FrozenSet[str]]:
        return None



# -*- coding: utf-8 -*-
__all__ = ('ClangAnalyser',)

from typing import Any, List, Mapping, Sequence
import json
import os

import dockerblade as _dockerblade
from loguru import logger

from .analysis import ClangFunction, ClangStatement
from ..analyser import Analyser
from ..analysis import Analysis
from ..core import FileLocationRange
from ..container import ProjectContainer
from ..exceptions import KaskaraException
from ..functions import ProgramFunctions
from ..loops import ProgramLoops
from ..project import Project
from ..statements import ProgramStatements
from ..util import abs_to_rel_flocrange


class ClangAnalyser(Analyser):
    def analyse(self, project: Project) -> Analysis:
        logger.debug(f"analysing Clang project: {project}")
        with project.provision() as container:
            return self._analyse_container(container)

    def _analyse_container(self, container: ProjectContainer) -> Analysis:
        loops = self._find_loops(container)
        functions = self._find_functions(container)
        statements = self._find_statements(container)
        insertions = statements.insertions()
        return Analysis(project=container.project,
                        loops=loops,
                        functions=functions,
                        statements=statements,
                        insertions=insertions)

    def _find_statements(self,
                         container: ProjectContainer
                         ) -> ProgramStatements:
        project = container.project
        logger.debug(f'finding statements for project: {project}')

        workdir = project.directory
        command_args = ['/opt/kaskara/scripts/kaskara-statement-finder']
        command_args += project.files
        command = ' '.join(command_args)
        output_filename = os.path.join(workdir, 'statements.json')
        logger.debug(f'executing statement finder [{workdir}]: {command}')
        try:
            container.shell.check_call(command, cwd=workdir)
        except _dockerblade.CalledProcessError as err:
            msg = f'statement finder failed with code {err.returncode}'
            logger.exception(msg)
            if not project.ignore_errors:
                raise KaskaraException(msg)

        logger.debug(f'reading results from file: {output_filename}')
        file_contents = container.files.read(output_filename)
        jsn: Sequence[Mapping[str, Any]] = json.loads(file_contents)
        statements = \
            ProgramStatements([ClangStatement.from_dict(project, d) for d in jsn])  # noqa
        logger.debug(f'finished reading results')
        return statements

    def _find_loops(self, container: ProjectContainer) -> ProgramLoops:
        project = container.project
        workdir = project.directory
        command_args = ['/opt/kaskara/scripts/kaskara-loop-finder']
        command_args += project.files
        command = ' '.join(command_args)
        output_filename = os.path.join(workdir, 'loops.json')
        logger.debug(f'executing loop finder [{workdir}]: {command}')
        try:
            container.shell.check_call(command, cwd=workdir)
        except _dockerblade.CalledProcessError as err:
            msg = f'loop finder failed with code {err.returncode}'
            logger.exception(msg)
            if not project.ignore_errors:
                raise KaskaraException(msg)

        logger.debug(f'reading results from file: {output_filename}')
        file_contents = container.files.read(output_filename)
        return self._read_loops_from_file_contents(project, file_contents)

    def _read_loops_from_file_contents(self,
                                       project: Project,
                                       contents: str
                                       ) -> 'ProgramLoops':
        loop_bodies: List[FileLocationRange] = []
        jsn: Sequence[Mapping[str, str]] = json.loads(contents)
        for loop_info in jsn:
            loc = FileLocationRange.from_string(loop_info['body'])
            loc = abs_to_rel_flocrange(project.directory, loc)
            loop_bodies.append(loc)
        logger.debug(f'finished reading loop analysis results')
        return ProgramLoops.from_body_location_ranges(loop_bodies)

    def _find_functions(self, container: ProjectContainer) -> ProgramFunctions:
        project = container.project
        workdir = project.directory
        output_filename = os.path.join(workdir, 'functions.json')
        command_args = ['/opt/kaskara/scripts/kaskara-function-scanner']
        command_args += project.files
        command = ' '.join(command_args)

        logger.debug(f'executing function scanner [{workdir}]: {command}')
        try:
            container.shell.check_call(command, cwd=workdir)
        except _dockerblade.CalledProcessError as err:
            msg = f'function scanner failed with code {err.returncode}'
            logger.exception(msg)
            if not project.ignore_errors:
                raise KaskaraException(msg)

        logger.debug(f'reading results from file: {output_filename}')
        file_contents = container.files.read(output_filename)
        jsn = json.loads(file_contents)
        return ProgramFunctions(ClangFunction.from_dict(project, d) for d in jsn)  # noqa



# -*- coding: utf-8 -*-
__all__ = ('PythonAnalyser',)

from loguru import logger

from .functions import collect_functions
from .statements import collect_statements
from .loops import collect_loops
from ..analyser import Analyser
from ..analysis import Analysis
from ..container import ProjectContainer
from ..project import Project


class PythonAnalyser(Analyser):
    def analyse(self, project: Project) -> Analysis:
        logger.debug(f"analysing Python project: {project}")
        with project.provision() as container:
            return self._analyse_container(container)

    def _analyse_container(self, container: ProjectContainer) -> Analysis:
        functions = collect_functions(container)
        statements = collect_statements(container)
        loops = collect_loops(container)
        insertions = statements.insertions()
        return Analysis(project=container.project,
                        functions=functions,
                        statements=statements,
                        insertions=insertions,
                        loops=loops)



# -*- coding: utf-8 -*-
__all__ = ('InsertionPoint', 'ProgramInsertionPoints')

from typing import FrozenSet, Iterable, Iterator, Dict, List, Any, Optional
import attr

from loguru import logger

from .core import FileLocation, FileLine
from .exceptions import KaskaraException
from .util import abs_to_rel_floc, rel_to_abs_floc


@attr.s(frozen=True, slots=True, auto_attribs=True)
class InsertionPoint:
    location: FileLocation
    visible: Optional[FrozenSet[str]] = attr.ib(eq=False, hash=False)


class ProgramInsertionPoints(Iterable[InsertionPoint]):
    def __init__(self, contents: List[InsertionPoint]) -> None:
        self.__contents = contents

        # index by file
        self.__file_insertions: Dict[str, List[InsertionPoint]] = {}
        for ins in contents:
            filename = ins.location.filename
            if filename not in self.__file_insertions:
                self.__file_insertions[filename] = []
            self.__file_insertions[filename].append(ins)

    def __iter__(self) -> Iterator[InsertionPoint]:
        yield from self.__contents

    def in_file(self, filename: str) -> Iterator[InsertionPoint]:
        """Returns an iterator over all insertion points in a given file."""
        logger.debug("finding insertion points in file: %s", filename)
        yield from self.__file_insertions.get(filename, [])

    def at_line(self, line: FileLine) -> Iterator[InsertionPoint]:
        """
        Returns an iterator over all of the insertion points located at a
        given line.
        """
        logger.debug(f"finding insertion points at line: {line}")
        filename: str = line.filename
        line_num: int = line.num
        for ins in self.in_file(filename):
            if line_num == ins.location.line:
                logger.debug(f"found insertion point at line [{line}]: {ins}")
                yield ins



# -*- coding: utf-8 -*-
"""
This module provides functionality for discovering and describing the set of
functions contained within a program.
"""
__all__ = ('Function', 'ProgramFunctions')

from typing import List, Dict, Tuple, Optional, Iterator, Iterable, Any
from typing_extensions import final
import abc

from .core import FileLocationRange, FileLocation


class Function(abc.ABC):
    """Provides a concise description of a function definition.

    Attributes
    ----------
    name: str
        The name of the function.
    location: FileLocationRange
        The range of locations spanned by the function.
    body_location: FileLocationRange
        The range of locations spanned by the function body.
    return_type: Optional[str]
        The name of the return type for the function, if known.
    filename: str
        The name of the file to which the function definition belongs.
    """
    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def return_type(self) -> Optional[str]:
        ...

    @property
    @abc.abstractmethod
    def location(self) -> FileLocationRange:
        ...

    @property
    @abc.abstractmethod
    def body_location(self) -> FileLocationRange:
        ...

    @final
    @property
    def filename(self) -> str:
        return self.location.filename


class ProgramFunctions:
    """Represents the set of functions within an associated program."""
    def __init__(self, functions: Iterable[Function]) -> None:
        self.__length = 0
        self.__filename_to_functions: Dict[str, List[Function]] = {}
        for f in functions:
            if f.filename not in self.__filename_to_functions:
                self.__filename_to_functions[f.filename] = []
            self.__filename_to_functions[f.filename].append(f)
            self.__length += 1

    def encloses(self, location: FileLocation) -> Optional[Function]:
        """Returns the enclosing function, if any, for a given location."""
        for func in self.in_file(location.filename):
            if location in func.location:
                return func
        return None

    def in_file(self, filename: str) -> Iterator[Function]:
        """Returns an iterator over the functions defined in a given file."""
        yield from self.__filename_to_functions.get(filename, [])

    def __len__(self) -> int:
        """Returns a count of the number of functions in the program."""
        return self.__length

    def __iter__(self) -> Iterator[Function]:
        """Returns an iterator over the functions in the program."""
        for functions in self.__filename_to_functions.values():
            yield from functions



# -*- coding: utf-8 -*-
__all__ = ('ClangFunction', 'ClangStatement')

from typing import Any, FrozenSet, Mapping

import attr

from ..core import FileLocationRange
from ..functions import Function
from ..project import Project
from ..statements import Statement
from ..util import abs_to_rel_flocrange


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ClangFunction(Function):
    name: str
    location: FileLocationRange
    body_location: FileLocationRange
    return_type: str
    is_global: bool
    is_pure: bool

    @staticmethod
    def from_dict(project: Project, d: Mapping[str, Any]) -> 'ClangFunction':
        name = d['name']
        location = FileLocationRange.from_string(d['location'])
        location = abs_to_rel_flocrange(project.directory, location)
        body = FileLocationRange.from_string(d['body'])
        body = abs_to_rel_flocrange(project.directory, body)
        return_type = d['return-type']
        is_global = d['global']
        is_pure = d['pure']
        return ClangFunction(name=name,
                             location=location,
                             body_location=body,
                             return_type=return_type,
                             is_global=is_global,
                             is_pure=is_pure)


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ClangStatement(Statement):
    content: str
    canonical: str
    kind: str
    location: FileLocationRange
    reads: FrozenSet[str]
    writes: FrozenSet[str]
    visible: FrozenSet[str]
    declares: FrozenSet[str]
    live_before: FrozenSet[str]
    requires_syntax: FrozenSet[str]

    @staticmethod
    def from_dict(project: Project, d: Mapping[str, Any]) -> 'ClangStatement':
        location = FileLocationRange.from_string(d['location'])
        location = abs_to_rel_flocrange(project.directory, location)
        return ClangStatement(d['content'],
                              d['canonical'],
                              d['kind'],
                              location,
                              frozenset(d.get('reads', [])),
                              frozenset(d.get('writes', [])),
                              frozenset(d.get('visible', [])),
                              frozenset(d.get('decls', [])),
                              frozenset(d.get('live_before', [])),
                              frozenset(d.get('requires_syntax', [])))



# -*- coding: utf-8 -*-
__all__ = ('ProgramLoops',)

from typing import Iterable

import attr

from .core import FileLocation, FileLocationRange, FileLocationRangeSet


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ProgramLoops:
    """Maintains information about all loops within a program."""
    _covered_by_loop_bodies: FileLocationRangeSet

    @classmethod
    def from_body_location_ranges(cls,
                                  bodies: Iterable[FileLocationRange]
                                  ) -> 'ProgramLoops':
        return ProgramLoops(FileLocationRangeSet(bodies))

    def is_within_loop(self, location: FileLocation) -> bool:
        """Checks whether a given location is enclosed within a loop."""
        return self._covered_by_loop_bodies.contains(location)



# -*- coding: utf-8 -*-
__all__ = ('Analyser',)

import abc
from typing_extensions import final

from .analysis import Analysis
from .project import Project


class Analyser(abc.ABC):
    """Provides a means of analysing projects."""
    @abc.abstractmethod
    def analyse(self, project: Project) -> Analysis:
        """Performs an analysis of a given project."""
        ...

    @final
    def __call__(self, project: Project) -> Analysis:
        """Alias for :meth:`analyse`."""
        return self.analyse(project)



# -*- coding: utf-8 -*-
__all__ = ('ast_with_tokens', 'ast_location')

from typing import Sequence, Union
import os

from ..core import Location, FileLocationRange, LocationRange

from loguru import logger
import ast
import asttokens

from ..container import ProjectContainer


def ast_with_tokens(container: ProjectContainer,
                    filename: str
                    ) -> asttokens.ASTTokens:
    """Retrieves the AST (with tokens) for a given file in a container."""
    logger.debug(f'retrieving AST for file [{filename}] '
                 f'in project [{container.project}]')
    assert not os.path.isabs(filename)
    abs_filename = os.path.join(container.project.directory, filename)
    file_contents = container.files.read(abs_filename)
    return asttokens.ASTTokens(file_contents, filename=filename, parse=True)


def ast_location(atok: asttokens.ASTTokens,
                 node: Union[ast.AST, Sequence[ast.AST]]
                 ) -> FileLocationRange:
    """Determines the source location of a node in an AST."""
    # mypy doesn't know that AST nodes are dynamically instrumented to
    # include first_token and last_token
    if isinstance(node, list):
        start_location = ast_location(atok, node[0]).start
        end_location = ast_location(atok, node[-1]).stop
        location_range = LocationRange(start_location, end_location)
        return FileLocationRange(atok._filename, location_range)

    filename = atok._filename
    line_numbers = atok._line_numbers
    offset_start = node.first_token.startpos  # type: ignore
    offset_end = node.last_token.endpos  # type: ignore
    line_start, col_start = line_numbers.offset_to_line(offset_start)
    line_end, col_end = line_numbers.offset_to_line(offset_end)
    location_start = Location(line=line_start, column=col_start)
    location_end = Location(line=line_end, column=col_end)
    location_range = LocationRange(location_start, location_end)
    return FileLocationRange(filename, location_range)



# -*- coding: utf-8 -*-
__all__ = ('Project',)

import typing
from typing import FrozenSet, Iterator

import contextlib

import attr
import docker as _docker
import dockerblade as _dockerblade

from .container import ProjectContainer

KASKARA_IMAGE = 'squareslab/kaskara'


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Project:
    """Describes a project under analysis.

    Attributes
    ----------
    image: str
        The name of the Docker image for this project.
    directory: str
        The absolute path of the root directory that holds the source code
        for this project.
    files: FrozenSet[str]
        The set of source code files for this project.
    ignore_errors: bool
        Indicates whether or not the analysis should proceed in the face of
        errors. If set to :code:`True`, the analysis will return partial
        results; if set to :code:`False`, an exception will be thrown instead.
    """
    _dockerblade: _dockerblade.DockerDaemon
    image: str
    directory: str
    files: FrozenSet[str]
    ignore_errors: bool = attr.ib(default=True)

    @contextlib.contextmanager
    def provision(self) -> Iterator[ProjectContainer]:
        """Provisions a Docker container for the project."""
        create = self._dockerblade.client.containers.create
        launch = self._dockerblade.client.containers.run
        with contextlib.ExitStack() as stack:
            docker_kaskara = create(KASKARA_IMAGE)
            stack.callback(docker_kaskara.remove, force=True)

            docker_project = launch(self.image, '/bin/sh',
                                    stdin_open=True,
                                    volumes_from=[docker_kaskara.id],
                                    detach=True)
            stack.callback(docker_project.remove, force=True)

            dockerblade = self._dockerblade.attach(docker_project.id)
            yield ProjectContainer(project=self, dockerblade=dockerblade)



# -*- coding: utf-8 -*-
import os
import re
from glob import glob
from setuptools import setup, find_packages

MODULE_FILE = \
    os.path.join(os.path.dirname(__file__), 'python/kaskara/version.py')
VERSION_REGEX = \
    r"__version__\s+=\s+['|\"](.*)['|\"]\s*"

with open(MODULE_FILE, 'r') as f:
    contents = f.read()
    VERSION = re.search(VERSION_REGEX, contents).group(1)

setup(version=VERSION)



# -*- coding: utf-8 -*-
import os

from .core import FileLocationRange, FileLocation


def abs_to_rel_filename(prefix: str, filename: str) -> str:
    if prefix[-1] != '/':
        prefix = prefix + '/'
    assert filename.startswith(prefix)
    return filename[len(prefix):]


def abs_to_rel_floc(prefix: str, l: FileLocation) -> FileLocation:
    return FileLocation(abs_to_rel_filename(prefix, l.filename),
                        l.location)


def rel_to_abs_floc(prefix: str, l: FileLocation) -> FileLocation:
    return FileLocation(os.path.join(prefix, l.filename), l.location)


def abs_to_rel_flocrange(prefix: str,
                         l: FileLocationRange
                         ) -> FileLocationRange:
    return FileLocationRange(abs_to_rel_filename(prefix, l.filename),
                             l.location_range)


def rel_to_abs_flocrange(prefix: str,
                         l: FileLocationRange
                         ) -> FileLocationRange:
    return FileLocationRange(os.path.join(prefix, l.filename),
                             l.location_range)



# -*- coding: utf-8 -*-
__all__ = ('Statement', 'ProgramStatements')

from typing import FrozenSet, Dict, List, Iterable, Iterator, Optional
import abc

from loguru import logger

from .core import FileLocationRange, FileLocation, FileLine
from .insertions import ProgramInsertionPoints, InsertionPoint


class Statement(abc.ABC):
    """Provides a description of a program statement.

    Attributes
    ----------
    kind: str
        The name of the kind of statement.
    content: str
        The original source code for this statement.
    canonical: str
        The canonical form of the source code for this statement.
    location: FileLocationRange
        The range of locations spanned by this statement.
    visible: FrozenSet[str], optional
        The set of visible symbols at this statement, if known.
    """
    @property
    @abc.abstractmethod
    def kind(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def content(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def canonical(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def visible(self) -> Optional[FrozenSet[str]]:
        ...

    @property
    @abc.abstractmethod
    def location(self) -> FileLocationRange:
        ...


class ProgramStatements:
    def __init__(self, statements: Iterable[Statement]) -> None:
        self.__statements = list(statements)
        logger.debug("indexing statements by file")
        self.__file_to_statements: Dict[str, List[Statement]] = {}
        for statement in statements:
            filename = statement.location.filename
            if filename not in self.__file_to_statements:
                self.__file_to_statements[filename] = []
            self.__file_to_statements[filename].append(statement)
        summary = '\n'.join(f"  {fn}: {len(stmts)} statements" for (fn, stmts)
                            in self.__file_to_statements.items())
        logger.debug(f'indexed statements by file:\n{summary}')

    def __iter__(self) -> Iterator[Statement]:
        yield from self.__statements

    def in_file(self, fn: str) -> Iterator[Statement]:
        """Returns an iterator over the statements belonging to a file."""
        yield from self.__file_to_statements.get(fn, [])

    def at_line(self, line: FileLine) -> Iterator[Statement]:
        """Returns an iterator over the statements located at a given line."""
        num = line.num
        for stmt in self.in_file(line.filename):
            if stmt.location.start.line == num:
                yield stmt

    def insertions(self) -> ProgramInsertionPoints:
        logger.debug('computing insertion points')
        points: List[InsertionPoint] = []
        for stmt in self:
            location = FileLocation(stmt.location.filename,
                                    stmt.location.stop)
            point = InsertionPoint(location, stmt.visible)

            # FIXME do not insert after a return

            points.append(point)
        db = ProgramInsertionPoints(points)
        logger.debug('computed insertion points')
        return db
